/* 
 * Tenshi no Hashigo for X
 * Copyright (c) 2001 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: $
 */
#include <stdio.h>
#include <stdlib.h>

#include "xladder.h"

#define IMG_RAW 0
#define IMG_LZ  1

#define RBUFSZ    (0xfff + 1)
#define LEN       (0xf + 3)

static void
decode_lz77(FILE *fp, u_char * data, int osize, int esize)
{
    char rbuf[RBUFSZ + LEN - 1];
    u_int ch, pos, c1, c2, offset, count, i;
    u_int bits = 0;
    int tpos = 0;

    memset(rbuf, 0, RBUFSZ - LEN);
    pos = RBUFSZ - LEN;

    while (!feof(fp)) {
	/* ե饰 (1B) ϲ̤鸫Ƥ */
	if (!((bits >>= 1) & 0x100)) {
	    ch = fgetc(fp);
	    bits = ch | 0xff00;
	}
	/* ӥåȤΩäƤ٥ 1B */
	if (bits & 1) {
	    ch = fgetc(fp);
	    rbuf[pos = pos++ & (RBUFSZ - 1)] = ch;
	    data[tpos] = ch;
	    if (++tpos > esize)
		return;
	} else {
	    /* 2B (ab cd) ǰ cab Ĺ d + 3 */
	    c1 = fgetc(fp);
	    c2 = fgetc(fp);
	    offset = c1 | ((c2 & 0xf0) << 4);
	    count = (c2 & 0x0f) + 3;
	    for (i = 0; i < count; i++) {
		ch = rbuf[(offset + i) & (RBUFSZ - 1)];
		rbuf[pos = pos++ & (RBUFSZ - 1)] = ch;
		data[tpos] = ch;
		if (++tpos > esize)
		    return;
	    }
	}
    }
}

static void
output_ppm(u_char * data, const char *name, int width, int height)
{
    FILE *fp;
    int j, k;

    fp = fopen(name, "w");
    fprintf(fp, "P6\n");
    fprintf(fp, "# CREATOR: Tenshi no Hashigo for X\n");
    fprintf(fp, "%d %d\n", width, height);
    fprintf(fp, "255\n");

#if 1
    for (j = height - 1; j >= 0; j--) {
	int p = (width * 3 + width % 4) * j;

	for (k = 0; k < width * 3; k += 3) {
	    fputc(data[p + k + 2], fp);
	    fputc(data[p + k + 1], fp);
	    fputc(data[p + k], fp);
	}
    }
#else
    for (j = 0; j < arc->item[i].esize; j++) {
	fputc(data[j], fp);
    }
#endif
    fclose(fp);
}

void
get_image(Archive *arc, const char *name)
{
    int i;
    u_char *data;
    int type;

    if (arc->type != ARC_BMP) {
	fprintf(stderr, "֤ǤϤޤ(%s)\n", name);
	return;
    }

    for (i = 0; i < arc->num; i++) {
	if (!strcmp(arc->item[i].name, name))
	    break;
    }
    if (i == arc->num) {
	fprintf(stderr, "Ĥޤ(%s)\n", name);
	return;;
    }

    if (arc->item[i].osize == arc->item[i].esize) {
	type = IMG_RAW;
	fprintf(stderr, "%s: raw image\n", name);
    } else if (name[0] == '0' || name[0] == 'm') {
	type = IMG_LZ;
	fprintf(stderr, "%s: lz77 image\n", name);
    }

    data = (u_char *) malloc(arc->item[i].esize);

    fseek(arc->fp, arc->item[i].pos, SEEK_SET);
    if (type == IMG_RAW) {
	fread(data, arc->item[i].esize, 1, arc->fp);
    } else {
	decode_lz77(arc->fp, data, arc->item[i].osize, arc->item[i].esize);
    }

    output_ppm(data, name, arc->item[i].width, arc->item[i].height);

    free(data);

    return;
}
